/* SPDX-License-Identifier: LGPL-2.1-or-later */
/*
 * Copyright (C) 2020, Google Inc.
 *
 * tracepoints.cpp - Tracepoints with lttng
 */
#define TRACEPOINT_CREATE_PROBES
#define TRACEPOINT_DEFINE

#include "libcamera/internal/tracepoints.h"
